;D16RUN.ASM	4-Aug-2005	Loren Blaney	loren_blaney@idcomm.com
;Run-time support code for D16XPLX (optimizing version) on the D16/M
;
;REVISIONS:
;4-Aug-2005: Changed d16x.tbl to d16.tbl (use corrected indexed opcodes)
;
;CODING CONVENTIONS:
;Subroutines do not destroy the contents of any registers. If a register is
; altered, this is stated as part of the output of the subroutine.
;
;Intrinsics assume that "the top of stack" (TOS) is in AC, with any additional
; arguments on the actual hardware stack. Intrinsics used as functions return
; their value in AC. Intrinsics may alter AC and IY, but not IX.
;
sim	equ	1		;1=simulator; 0=actual computer hardware
	if	sim
data	 setl	3000h		;start of data memory space
	else
data	 setl	8000h		;start of data memory space
	endi
code	setl	0		;start of code section of memory

	cpu	"d16.tbl"	;assembler's required processor table
	hof	"mot16"		;use Motorola hex format
	wdln	2		;word length = 2 bytes (not 1)
	org	code

	if	sim
	 lds	05FFDh		;initialize stack pointer to top of memory
heaphi	 equ	05800h		;top of heap (highest addr + 1)
	else
	 lds	0FFFDh		;initialize stack pointer to top of memory
heaphi	 equ	0F000h		;top of heap (highest addr + 1)
	endi

;Jump table to entry points (convenient for starting Debug Monitor)
	jmp	start		;XPL entry point
	jmp	dMon		;Debug Monitor

;-------------------------------------------------------------------------------

code	setl	$		;switch to data section of memory
	org	data
temp	dwm	0		;very tempory scratch location
stkptr	dwm	0		;stack pointer (SP) when XPL code started
hp	dwm	0		;heap pointer (base of XPL's free memory)
nowDev	dwm	0		;current I/O device number (0..8)
errNum	dwm	0		;error code number (0..15)
traps	dwm	0		;bit array: run-time error trap flags
remain	dwm	0		;remainder from last integer divide
rerunF	dwm	0		;rerun flag, set by Restart, intrinsic 6

;Display vector. These point to the base of variables in the heap, one for each
; level of static nesting. In practice, base3 and above are rarely used.
base0	dwm	0		;base of global variables
base1	dwm	0		;base of local variables at level 1
base2	dwm	0		;etc.
base3	dwm	0
base4	dwm	0
base5	dwm	0
base6	dwm	0
base7	dwm	0

;Some ASCII control characters:
bel	equ	07h		;bell
bs	equ	08h		;backspace
cr	equ	0Dh		;carriage return
lf	equ	0Ah		;line feed
eof	equ	1ah		;end of file
tsp	equ	" "+80h		;string terminating space character (printed)
data	setl	$
	org	code

;===============================================================================
;				ENTRY POINT
;===============================================================================
;
start	clr			;no errors yet
	sta	(errNum)
	set			;enable trapping for all error types
	sta	(traps)
	lda	heapLo
	sta	(hp)
	sta	(base0)		;set level 0 display vector
	csr	setInt		;initialize intrinsics

	csr	startXpl	;run the compiled XPL program
	lda	(errNum)	;did a run-time error occur?
	joz	start90		;skip if not
start80
	clr			;display an error message
	sta	(nowDev)	;output to the console
	lda	em0
	csr	textP
	lda	(errNum)	;display error number
	csr	intOut
	ldy	(errNum)	;index to appropriate message
	lda	(iy+errMsg-1)
	csr	textP
	csr	crlf		;new line
start90
	jmp	dMon		;go to Debug Monitor

;-------------------------------------------------------------------------------
;Set up and start the XPL program.
;
startXpl
	sts	(stkptr)	;save stack pointer (so we can return from
	jmp	progrm		; anywhere) then jump to comiled XPL code

;===============================================================================
;Run-time error handler. AC contains the error number.
;
error	psa			;save register(s)
	sta	(errNum)	;save the error number
	sta	(errCtr)

	lda	(traps)		;see if this error is being trapped
	lsr			;shift corresponding trap bit into carry
err10	lsr			;bit number = error number
	dsz	(errCtr)	;(this is NOT the way it's done on the PC)
	 jmp	err10

	jnc	err90		;if bit is clear then don't trap, just return
	lds	(stkPtr)	;else trap error and exit XPL program
	rtn			;return from anywhere
err90
	poa			;restore register(s)
	rtn

code	setl	$		;switch to data section of memory
	org	data
errCtr	dwm	0		;loop counter

;Error messages:
em0	dfb	bel, cr, lf, "RUN-TIME ERROR", tsp

;WARNING!!! WARNING!!! WARNING!!! DFB's are unbelievably TREACHEROUS!!!
; The assembler DOES NOT handle them correctly. The listing shows one thing
; and the hex file LOADS ANOTHER. They MUST contain an even number of bytes
; to work properly. The ALGN directive DOES NOT WORK. Separating DFB's with
; DWM's does not work.

errMsg	dwm	em1		;table of message pointers
	dwm	em2
	dwm	em3
	dwm	em4
	dwm	em5
	dwm	em6
	dwm	em7
	dwm	em8
	dwm	em9
	dwm	em10
	dwm	em11
	dwm	em12
	dwm	em13
	dwm	em14
	dwm	em15

em1	dfb	": DIV BY 0", tsp, tsp
em2	dfb	": OUT OF MEMORY", tsp
em3	dfb	": I/O", tsp
em4	dfb	": BAD OPCODE", tsp, tsp
em5	dfb	": BAD INTRINSIC", tsp
em6	dfb	": DIV BY 0.0", tsp, tsp
em7	dfb	": OVERFLOW", tsp, tsp
em8	dfb	": UNDERFLOW", tsp
em9	dfb	": FIX OVERFLOW", tsp, tsp
em10	dfb	": SQRT < 0.0", tsp, tsp
em11	dfb	": LOG <= 0.0", tsp, tsp
em12	dfb	": EXP OVERFLOW", tsp, tsp
em13	dfb	": OVERFLOW", tsp, tsp
em14	dfb	": ATAN2(0.0,0.0)", tsp, tsp
em15	dfb	": VIDEO MODE", tsp, tsp
data	setl	$
	org	code

;===============================================================================
;                               INTRINSICS
;===============================================================================
;Start-up initialization for intrinsics.
;
setInt	csr	resBuf		;initialize keyboard buffer
	jmp	ranInt		;reset random number generator seed

;-------------------------------------------------------------------------------
;Intrinsic to return the absolute value of top of stack.
; VAL:= ABS(VAL);
;
intr0	jpl	abs10
	neg
abs10	rtn

;-------------------------------------------------------------------------------
;Intrinsic to generate a random number between 0 and (top of stack)-1.
; If TOS = 0 then initialize seed for repeatable sequence.
; If TOS < 0 then randomize and return RAN(-TOS).
; VAL:= RAN(10);	\random number from 0 through 9
;
code	setl	$		;switch to data section of memory
	org	data
ranK	dwm	2537
ranL	dwm	5149
ranM	dwm	7026

modK	equ	33049		;prime numbers that give good results
modL	equ	32909
modM	equ	32771
data	setl	$
	org	code

intr1	jnz	ran01		;range is in AC, if AC=0 then
	lda	2537		; set seeds for a repeatable sequence
	sta	(ranK)
	lda	5149
	sta	(ranL)
	lda	7026
	sta	(ranM)
	rtn
ran01
	jpl	ran05
	neg
	csr	ranInt		;randomize seed
ran05
	sta	(temp)		;save divisor

	lda	(ranK)		;ranK:= rem((ranK+ranK) / modK);
	shl
	sub	modK
	jnc	ran10
	add	modK
ran10	sta	(ranK)

	lda	(ranL)		;ranL:= rem((ranL+ranL) / modL);
	shl
	sub	modL
	jnc	ran20
	add	modL
ran20	sta	(ranL)

	add	(ranK)		;ranM:= rem((ranK+ranL+ranM) / modM);
	add	(ranM)
ran30	sub	modM
	jnc	ran30
	add	modM
	sta	(ranM)

	may
	lda	(temp)		;get divisor
	csr	div		;AC:= IY / AC; remainder is returned in IY
	mya			;return remainder in AC
	rtn

;-------------------------------------------------------------------------------
;Initialize random number generator.
;
ranInt	psa

	lda	(spinner)	;get random seed  ??? (put into Xmit wait loop)
	ora	4		;make sure it's not zero
	sta	(ranK)		;set seed
	poa
	rtn

code	setl	$
	org	data
spinner	dwm	0
data	setl	$
	org	code

;-------------------------------------------------------------------------------
;Intrinsic to set the seed for the random number generator.
; SETSEED(Seed);
;
intr79	sta	(ranK)
	lda	5149		;set seeds for a repeatable sequence
	sta	(ranL)
	lda	7026
	sta	(ranM)
	rtn

;-------------------------------------------------------------------------------
;Intrinsic to return the remainder of the most recent division. The argument
; is an expression whose result is thrown away. This expression can contain
; a division, or simply be a zero to get the result of an earlier division.
; VAL:= REM(VAL/10);
;
intr2	lda	(remain)	;get last remainder
	rtn

;-------------------------------------------------------------------------------
;Intrinsic to reserve bytes on the heap and return the address of the
; reserved space. This is the trick by which XPL handles dynamic storage
; for arrays. Since the space is reserved in the heap allocation of a
; procedure, the array will disappear when the procedure is exited.
; ARRAY:= RESERVE(10 *INTSIZE);
;
intr3	ldy	(hp)		;save heap pointer
	add	(hp)		;add in number of words ?? to reserve
	sta	(hp)
	joc	res80		;error if heap wraps around
	sub	heaphi		;test for overflow
	joc	res90		;skip if ok
res80	lda	2		;run-time error # 2: out of memory
	csr	error
res90	mya			;return address of reserved space
	rtn

;-------------------------------------------------------------------------------
;Intrinsic to swap the two low bytes of top of stack.
; VAL:= SWAP(VAL);
; (The compiler depends on this not altering the IY register. See 'swap'.)
;
intr4	sta	(temp)
	and	0FF00h		;get high byte
	add	(temp)		;shift high byte left into carry
	rol			;rotate 8 times
	rol
	rol
	rol
	rol
	rol
	rol
	rol
	rtn

;-------------------------------------------------------------------------------
;Intrinsic to extend the sign of the low byte into the high bytes.
; VAL:= EXTEND(VAL);
; (The compiler depends on this not altering the IY register. See 'ext'.)
;
intr5	psa
	and	0080h		;test sign of low byte
	joz	ext10		;jump if positive byte
	poa
	ora	0FF00h		;set high bits
	rtn
ext10
	poa
	and	00FFh		;clear high bits
	rtn

;-------------------------------------------------------------------------------
;Intrinsic to restart the current program.
; RESTART;
;
intr6	set			;set rerun flag
	sta	(rerunF)
	jmp	start

;-------------------------------------------------------------------------------
;Intrinsic to input a byte and return it in AC.
; CH:= CHIN(DEV);
;
intr7	sta	(nowDev)	;set device number
	csr	chIn		;get byte
	and	00FFh		;zero high bytes
	rtn

;-------------------------------------------------------------------------------
;Intrinsic to output the byte in AC.
; CHOUT(DEV, CH);
;
intr8	ldy	(sp+1)		;set output device number
	sty	(nowDev)
	poy			;get return address
	ins			;discard argument from stack
	psy			;restore return address
	jmp	chOut		;output it and return

;-------------------------------------------------------------------------------
;Intrinsic to output a carriage return and a line feed.
; CRLF(DEV);
;
intr9	sta	(nowDev)	;set device number
	jmp	crlf

;-------------------------------------------------------------------------------
;Intrinsic to input ASCII digits and convert them to a signed, decimal, value
; that is returned in AC. The number must be terminated with a non-digit
; character (such as CR). This routine skips any leading non-digit characters.
; This reads in either an unsigned number in the range 0 to 65535, or a signed
; number in the range -32768 to 32767. There is no error detected for overflows
; outside these ranges.
;  VAL:= INTIN(DEV);
;
intr10	sta	(nowDev)	;set device number

	clr			;initialize
	sta	(regA)		;decimal value accumulator
	sta	(haveNum)	;flag
	sta	(isMinus)	;flag

di10	csr	chIn		;get byte
	sub	"0"		;convert ASCII dec character to binary
	sta	(temp)		;shift this dec digit into regA
	sub	9+1
	jnc	di30		;jump if not decimal digit char
	isz	(haveNum)	;indicate a decimal character has been found
				; (never skips)
	lda	(regA)		;regA:= regA*10 + digit
	shl
	shl
	add	(regA)
	shl
	add	(temp)
	sta	(regA)
	jmp	di10		;loop back for next character
di30				;come here if non-decimal character
	lda	(haveNum)	;has a decimal character been processed?
	jnz	di40		;jump if so

	lda	(temp)		;is char a minus sign?
	add	"0" - "-"
	jnz	di10		;loop back if not--ignore leading garbage
	 isz	(isMinus)	;set minus flag (never skips)
	jmp	di10		;loop back for next character
di40
	lda	(isMinus)	;is the number supposed to be negative?
	lsr			;(note that -- = +)
	lda	(regA)
	jnc	di90		;jump if not
	 neg
di90	rtn			;return with dec value in AC

code	setl	$
	org	data
isMinus	dwm	0		;bit 0 is set if minus sign was read in
data	setl	$
	org	code

;-------------------------------------------------------------------------------
;Intrinsic to output the signed integer on top of stack (AC).
; INTOUT(DEV, VAL);
;
intr11	ldy	(sp+1)		;get device number
	sty	(nowDev)
	poy			;get return address
	ins			;discard argument from stack
	psy			;restore return address
;fall into intOut routine...

;-------------------------------------------------------------------------------
;Convert the value in AC to an ASCII, signed, decimal number, and output it to
; outDev. AC and IY are destroyed.
;
intOut	jpl	into10
	neg			;if the number is negative then make it positive
	psa			;output a minus sign
	lda	"-"
	csr	chOut
	poa
into10
into15	may
	lda	10
	csr	div		;AC:= IY / AC; remainder is returned in IY
	psy			;save remainder on stack
	joz	into20		;if the quotient > 0 then output it as
	 csr	into15		; a decimal number
into20	poa			;retrieve remainder
	ora	"0"		;convert it to ASCII
	jmp	chOut		;output it and return

;-------------------------------------------------------------------------------
;Intrinsic to output a text string. The starting address is top of stack (AC).
; The string is terminated by a character with bit 7 set.
;  TEXT(DEV, "HELLO WORLD");
;
intr12	ldy	(sp+1)		;get device number
	sty	(nowDev)
	poy			;get return address
	ins			;discard argument from stack
	psy			;restore return address

	may			;IY:= address of string
txto20	lda	(iy+0)		;get character
	and	0080h		;exit loop if MSB is set
	jnz	txto35
	lda	(iy+0)
	iny
	csr	chOut		;output character
	jmp	txto20
txto35
	lda	(iy+0)		;get character again
	and	007Fh		;mask off MSB
	jmp	chOut		;output last character and return

;-------------------------------------------------------------------------------
;Intrinsic to open an input device.
; OPENI(DEV);
;
intr13	sta	(nowDev)	;get device number
	jmp	openI		;go do it and return

;-------------------------------------------------------------------------------
;Intrinsic to open an output device.
; OPENO(DEV);
;
intr14	sta	(nowDev)	;get device number
	jmp	openO		;go do it and return

;-------------------------------------------------------------------------------
;Intrinsic to close an output device.
; CLOSE(DEV);
;
intr15	sta	(nowDev)	;get device number
	jmp	close		;go do it and return

;-------------------------------------------------------------------------------
;Intrinsic to abort the program.
; ABORT;
;
intr16	lds	(stkptr)
	rtn

;-------------------------------------------------------------------------------
;Intrinsic to set the error trap flags.
; TRAP($FFFF);
;
intr17	sta	(traps)
	rtn

;-------------------------------------------------------------------------------
;Intrinsic to tell a user how many bytes of heap space he has left. Of
; course he may not reserve all of it, since he must leave a working
; heap for subsequent procedure calls etc. Since only the user knows
; how much this might be, it is left to him to decide how much he
; actually has free to play with.
; VAL:= FREE;
;
intr18	lda	heaphi
	sub	(hp)
	rtn

;-------------------------------------------------------------------------------
;Intrinsic to return the rerun flag.
; VAL:= RERUN;
;
intr19	lda	(rerunF)
	rtn

;-------------------------------------------------------------------------------
;Intrinsic to return the current value of the heap pointer. This intrinsic
; is used for saving the heap pointer for later restoration with SETHP.
; The user had better have a good idea of the funtioning of XPL before
; dinging with the heap pointer or he will surely bomb himself!
; VAL:= GETHP;
;
intr20	lda	(hp)
	rtn

;-------------------------------------------------------------------------------
;Intrinsic to set the heap pointer--a very dangerous thing to do! See
; GETHP's comments.
; SETHP(VAL);
;
intr21	sta	(hp)
	rtn

;-------------------------------------------------------------------------------
;Intrinsic to return the run-time error number and then clear it.
; VAL:= GETERR;
;
intr22	ldy	(errNum)
	clr
	sta	(errNum)
	mya
	rtn

;-------------------------------------------------------------------------------
;Intrinsic to set the video cursor to a specified position.
; CURSOR(X, Y);
;
intr23	ldy	(sp+1)		;get X coordinate
;	sty	(CursorX)
	poy			;get return address
	ins			;discard argument from stack
	psy			;restore return address
	rtn

;-------------------------------------------------------------------------------
;Intrinsic to set the rerun flag directly.
; SETRUN;
;
intr25	sta	(rerunF)
	rtn

;-------------------------------------------------------------------------------
;Intrinsic to input ASCII digits and convert them to a 16-bit hex value
; which is returned in AC.
; VAL:= HEXIN(DEV);
;
intr26	sta	(nowDev)	;set device number
	jmp	hexIn		;go do it

;-------------------------------------------------------------------------------
;Intrinsic to output the top of stack (AC) in hex ASCII.
; HEXOUT(DEV, VAL);
;
intr27	ldy	(sp+1)		;get device number
	sty	(nowDev)
	poy			;get return address
	ins			;discard argument from stack
	psy			;restore return address
	jmp	hex4out		;output AC in hex ASCII and return

;===============================================================================
;Multiply two 16-bit, signed numbers
; AC:= AC * IY
; This uses an early-out algorithm.
; The IY register is destroyed.
;
;	  Ŀ   Ŀ
;	< IX       IY >
;	     
;	  Ŀ
;	+  temp 
;	  
;
mulDo	psx			;save IX register

	max			;put multiplicand into IX
	sty	(temp)		;put smaller value into IY and larger into IX
	sub	(temp)		;if (unsigned) AC < IY then swap AC and IY
	jnc	mul05		;jump if not
	mxa			;swap IX & IY
	may
	ldx	(temp)
mul05
	clr			;clear product register (temp); temp:= 0
	sta	(temp)
	jmp	mul30		;enter main loop

mul10	mxa			;add multiplicand to temp; temp:= temp + IX
	add	(temp)
	sta	(temp)

mul20	mxa			;shift multiplicand left; << IX
	shl
	max
mul30
	mya			;shift least significant bit of multiplier
	lsr			; into carry; IY >>
	may
	joc	mul10		;jump if it's a 1--go add multiplicand
	jnz	mul20		;stop looping when there are no more 1's in IY

	lda	(temp)		;return product in AC
	pox			;restore IX register
	rtn

;-------------------------------------------------------------------------------
;Signed 16-bit divide
; AC:= AC/IY; remain:= remainder (also IY:= remainder)
;
divDo	sty	(temp)		;save divisor
	psa			;save dividend
	xor	(temp)
	sta	(sign)		;if MSB is set then quotient is negative
	poa			;restore dividend

	jpl	div03		;IY:= absolute value of dividend
	 neg
div03	may
	lda	(temp)		;AC:= divisor
	jpl	div05		;take absolute value
	 neg
div05
	csr	div		;AC:= IY / AC; remainder is returned in IY

	psa			;save quotient
	lda	(sign)		;determine sign of result
	shl			;shift sign bit into carry
	poa			;get quotient
	jnc	div07		;make it negative if necessary
	 neg
div07
	sty	(remain)	;set up remainder (always positive)
	rtn

code	setl	$
	org	data
sign	dwm	0
divCtr	dwm	0
data	setl	$
	org	code

;-------------------------------------------------------------------------------
;Unsigned 16/16-bit divide
; AC:= IY / AC; remainder is returned in IY
;
;	Ŀ  Ŀ  Ŀ  Ŀ
;	  IY  <ĳC<ĳ  IX  <ĳC
;	      
;	Ŀ
;     -  temp 
;	
;
div	jnz	div08		;skip if ok
	 lda	1		;flag div-by-0 error
	 csr	error		;(use 1 instead of 0)
	 lda	7FFFh		;best answer is: quotient = infinity
	 ldy	0		;remainder = 0
	 rtn
div08
	sta	(temp)		;set up the divisor; temp:= AC

	mya			;get dividend (numerator)
	joz	div99		;jump if 0; quotient = 0; remainder = 0

	psx			;preserve IX register
	ldx	17		;initialize loop counter
	com			;use inverted logic so 1's get shfited in
div09	shl			;quickly left justify dividend (numerator)
	dex			;adjust loop counter
	joc	div09		;loop
	com			;reinvert logic
	stx	(divCtr)	;set loop counter
	max			;set IX with justified dividend
	lda	1		;initialize IY
	jmp	div12

div10	max
	mya			;<< IY:IX
	rol
div12	may

	sub	(temp)		;IY - temp
	joc	div20		;skip if temp > IY (temp does not go into IY)
	 may			;else save the difference
div20
	mxa			;<< IX
	rol			;shift in carry bit (gets inverted later)

	dsz	(divCtr)	;loop for a total of 16 bits
	 jmp	div10

	com			;invert all the carry bits
	pox			;restore IX register
div99	rtn			;return with quotient in AC

;-------------------------------------------------------------------------------
;Return the sqare root of an the unsigned value in AC. AC:= sqrt(AC)
; Largest error is less than 0.6 [sqrt(2) -> 2 vs: 1.414...]
;
sqrt	joz	sqrt30		;sqrt(0) = 0
	lsr			;divide by 2
	psa			;zero regA
	clr
	sta	(regA)
	poa
sqrt10	isz	(regA)		;(never skips)
	sub	(regA)
	jpl	sqrt10
	lda	(regA)
sqrt30	rtn

;-------------------------------------------------------------------------------
;Routine to reserve integer and real arrays with any number of dimensions.
; This also sets the pointer to the heap space that is reserved.
; For example: Mkarray( (3, 5, 7, 11), 4, 2);
; The stack contains:
;  The number of items in each dimension (3, 5, 7, 11)
;  The number of dimensions (4)
; AC = the number of bytes in each item of the array (1=char, 2=int, 8=real)
; The ac and iy registers are altered by this routine (ix is preserved).
;
;	  stack			  heap
; sp0->	| retadr |   iy, hp0 ->	|pointer|		addresses
;	|  4     |	 hp1 ->	| array | <- hBase	increase
;	| 11     |		| space |		downward
;	|  7     |		|       |		    |
;	|  5     |		|       |		    V
;	|  3     | <- frame	|       |
;
code	setl	$		;switch to data section of memory
	org	data
retAdr	dwm	0		;return address
ixSave	dwm	0		;save location for critical registers
nWords	dwm	0		;number of words in each array item (1 or 4)
frame	dwm	0		;pointer to first argument pushed on the stack
levMax	dwm	0		;maximum level of recursion (last dimension)
items	dwm	0		;number of items in array
hBase	dwm	0		;base address of current dimension of array
data	setl	$
	org	code

mkarray	poy			;remove return address from stack so we can
	sty	(retAdr)	; easily access the arguments
	stx	(ixSave)	;save ix register

	inc			;convert bytes to words
	lsr			; 1 to 1, 2 to 1, and 8 to 4
	sta	(nWords)	;save number of words in each array item
	ldy	(hp)		;iy points to pointer to array
	add	(hp)		;allocate heap space for pointer to array (1, 4)
	sta	(hp)

	poa			;get number of dimensions
	dec			;convert to maximum index off of "frame" (levMax)

	sta	(frame)		;point to number of items in first dimension
	neg
	sta	(levMax)
	msa
	add	(frame)
	sta	(frame)

	ldx	0		;set index to point to size of first dimension
	csr	genAry		;reserve first dimension of array and store its
				; base address into location pointed to by ebx
	lds	(frame)		;clean up stack--discard arguments
	ins

	ldx	(ixSave)	;restore ix register
	lda	(retAdr)	;restore return address
	psa
	rtn

;Recursively called subroutine for setting up a multi-dimensional array.
;Reserve an array and store its base address into the location pointed to
; by the iy register.
;Inputs:
; ix: index (backward from frame) to size of each dimension
;     also equals level of recursion
; iy: address of pointer to (next dimension of) array

genAry	mxa			;get number of items in current dimension
	add	(frame)		;lda (ix+(frame))
	psx
	max
	lda	(ix+0)
	pox
	sta	(items)

;Reserve space on the heap

	lda	(hp)		;base of reserved heap space
	sta	(hBase)

	lda	(nWords)	;is this a real array?
	sub	4
	jnz	ga05		;skip if not
	lda	(items)		;multiply items by 4 words per real number
	shl
	joc	ga10		;check for memory overflow: error if > $FFFF
	shl
	joc	ga10		;check for memory overflow: error if > $FFFF
	jmp	ga06
ga05	 lda	(items)		;integer (and char) have one word per item
ga06	add	(hp)		;add number of words to reserve to heap pointer
	sta	(hp)
	joc	ga10		;check for memory overflow: error if > $FFFF
	lda	heaphi		;check for memory overflow
	sub	(hp)
	jnc	ga20		;skip if ok
ga10
	lda	2		;error 2: memory overflow
	csr	error
ga20
	lda	(hBase)		;store base address of reserved space into ptr
	sta	(iy+0)

	mxa			;is this the last dimension?
	sub	(levMax)
	joz	ga90		;jump if so--don't recurse any lower

;For each item in the current dimension, reserve space for the next dimension
	clr			;init index to first item in current dimension
ga30	sub	(items)		;is this ac index at the limit?
	joz	ga90		;jump if so--exit loop
	add	(items)		;restore ac index

	dex			;move ix index to next dim. size in the stack

	psa			;save ac index

	may			;also save it in iy
	lda	(nWords)	;is this a real array (vs. integer)?
	sub	4
	jnz	ga40		;skip if not
	mya			;else reserve space for 4 words
	shl			;multiply the ac index by 4
	shl
	jmp	ga45
ga40	 mya			;get ac index (multiplied by 1)
ga45	add	(hBase)		;add base address of reserved space to get the
	may			; address to put the pointer to next dimension
				; pass this address via iy
	lda	(items)		;save some things across subroutine call
	psa
	lda	(hBase)
	psa

	csr	genAry		;reserve next dimension of array and store its
				; base address into location pointed to by iy
	poa			;restore things
	sta	(hBase)
	poa
	sta	(items)
	poa

	inx			;restore ix index to original dimension in stack

	inc			;next item in first (or current) dimension
	jmp	ga30		;loop for each item in current dimension
ga90
	rtn

;###############################################################################
;                              Debug Monitor
;###############################################################################

dMon	clr			;initialize registers for neatness
	sta	(nowDev)	;all I/O is thru the console (device 0)
	sta	(pcreg)
	sta	(acreg)
	sta	(flreg)
	sta	(bpreg)
	sta	(opreg)
	sta	(op2reg)
	sts	(spreg)

;Order matches R command
code	setl	$
	org	data
pcreg	dwm	0		;program counter
acreg	dwm	0		;accumulator
flreg	dwm	0		;flag register bits: 2=IE, 1=OV, 0=CY
bpreg	dwm	0		;breakpoint location
opreg	dwm	0		;instruction at breakpoint
op2reg	dwm	0		;2nd word of instruction
spreg	dwm	0		;stack pointer
data	setl	$
	org	code

cmd00	csr	crlf
	lda	":"		;display (DMONic) prompt
	csr	chOut
	lda	">"
	csr	chOut

	csr	resBuf		;initialize keyboard buffer
	csr	bufKey		;get command letter
	and	0DFh		;shift to uppercase
	sta	(cmd)
	lda	cmdtbl
	sta	(ptr)

cmd10	lda	[ptr++]		;get command letter from table
	joz	terror		;jump if terminator (=0)
	sub	(cmd)
	joz	(ptr)		;jump (indirect) to handler
	isz	(ptr)		;move past jmp to handler (never skips)
	isz	(ptr)
	jmp	cmd10

cmdtbl	dwm	"D"		;dump memory starting at given address
	jmp	dump		;handler entry point
	dwm	"R"		;dump all registers
	jmp	regs
	dwm	"L"		;load file
	jmp	load
	dwm	"G"		;go execute at given address
	jmp	go
	dwm	"S"		;set memory contents at given address
	jmp	set
	dwm	"B"		;set (move) breakpoint to given address
	jmp	break
	dwm	0		;terminate table entries

code	setl	$
	org	data
cmd	dwm	0		;command letter
ptr	dwm	0		;pointer into cmdtbl
data	setl	$
	org	code

;-------------------------------------------------------------------------------
;Display error message and return to dispatch loop

terror	lda	"?"
	csr	chOut
	jmp	cmd00		;return to main dispatch loop

;-------------------------------------------------------------------------------
;D AAAA = Dump 32 words (4 lines) of memory starting at specified address.
; Each line is displayed in this format:
;  AAAA: XXXX XXXX XXXX XXXX  XXXX XXXX XXXX XXXX 
; If address (AAAA) is not given then resume following last dump address.
; Add ASCII dump eventually.
; Start second line on an address evenly divisible by 8 (eventually).

dump	csr	hexin		;get starting address
	sta	(addr)

	lda	4		;for 4 lines...
	sta	(j)
dmp10	lda	(addr)
	csr	hex4out
	lda	":"		;output ": "
	csr	chOut
	csr	spOut

	lda	8		;for 8 words per line...
	sta	(i)

dmp15	lda	[addr++]	;output word at addr
	csr	hex4out
	csr	spOut
	lda	(i)		;put an extra space in the center 
	sub	5
	jnz	dmp20
	 csr	spOut
dmp20
	dsz	(i)		;next word
	 jmp	dmp15

	csr	spOut
	lda	(addr)		;repeat dump in ASCII
	sub	8
	sta	(addr)

	lda	8		;for 8 words per line...
	sta	(i)
dmp30	lda	[addr]		;output character at addr
	csr	lsr8
	csr	charout
	lda	[addr++]	;output character at addr
	csr	charout
	dsz	(i)		;next word
	 jmp	dmp30

	csr	crlf
	dsz	(j)		;next line
	 jmp	dmp10
	jmp	cmd00		;return to main dispatch loop

;Display characters in the range 20h..7Eh, otherwise replace them with a dot.
charout	and	7Fh		;make it ASCII
	psa
	sub	20h
	sub	7Fh-20h
	poa			;get char
	joc	charo10		;jump if within range
	 lda	"."
charo10	jmp	chOut		;display char and return

code	setl	$
	org	data
addr	dwm	0		;pointer to address(es) to dump or load
i	dwm	0		;loop counter
j	dwm	0		;loop counter
data	setl	$
	org	code

;-------------------------------------------------------------------------------
;R = Dump all registers.
; PC:XXXX   AC:XXXX   CY:X   OV:X   IE:X   BP:XXXX   OP:XXXX XXXX   SP: XXXX
; BP is the breakpoint address, OP is the opcode at the breakpoint.

regs	lda	pcmsg
	csr	textZ
	lda	(pcreg)
	csr	hexout
	csr	sp3out

	lda	acmsg
	csr	textZ
	lda	(acreg)
	csr	hexout
	csr	sp3out

	lda	(flreg)
	sta	(flreg2)
	lda	cymsg
	csr	flagout
	lda	ovmsg
	csr	flagout
	lda	iemsg
	csr	flagout

	lda	bpmsg
	csr	textZ
	lda	(bpreg)
	csr	hexout
	csr	sp3out

	lda	opmsg
	csr	textZ
	lda	(opreg)
	csr	hexout
	csr	spOut
	lda	(op2reg)
	csr	hexout
	csr	sp3out

	lda	spmsg
	csr	textZ
	lda	(spreg)
	csr	hexout
	csr	sp3out
	jmp	cmd00		;return to main dispatch loop

;Display flag bits
flagout	csr	textZ
	lda	(flreg2)	;shift status bit into carry
	ror
	sta	(flreg2)
	lda	"0"		;assume status bit is clear
	adc	0		;change "0" to "1" if carry is set
	csr	chOut
	jmp	sp3out		;3 spaces then return

pcmsg	dfb	"PC:", 0
acmsg	dfb	"AC:", 0
cymsg	dfb	"CY:", 0
ovmsg	dfb	"OV:", 0
iemsg	dfb	"IE:", 0
bpmsg	dfb	"BP:", 0
opmsg	dfb	"OP:", 0
spmsg	dfb	"SP:", 0

code	setl	$
	org	data
flreg2	dwm	0		;flag register, working copy
data	setl	$
	org	code

;-------------------------------------------------------------------------------
;L = Load a file in MOT16 format.
; (Not position independent code thus no load address specified.)

load
ld10	csr	chIn		;skip to start of record
	sub	"S"
	jnz	ld10

	csr	chIn		;get the record type: 2=data; 8=EOF
	sta	(type)

	csr	hex2in		;get number of bytes in this record
	sub	4		;subtract 3 for the load addr + 1 for checksum
	lsr			;divide by 2 to get the number of words
	inc			;+ 1 to handle zero-length records
	sta	(words)

;Byte address range = 000000..01FFFF (= 64K words)
	csr	hex2in		;get the byte address to load
	lsr			;we only care about the LSB
	psf			; so save it in the carry flag
	csr	hex4in		;get rest of the byte address to load
	pof			;divide by 2 to get the word address
	ror
	sta	(addr)
	jmp	ld30

ld20	csr	hex4in		;read in a word of data
	sta	[addr++]	;store it
ld30	dsz	(words)		;loop for the number of words in this record
	 jmp	ld20

	csr	chIn		;read in but ignore the checksum (it's a pain
	csr	chIn		; to calculate a byte cksum with words)

	lda	(type)		;loop while record type = "2"
	sub	"2"
	joz	ld10

	csr	chIn		;eat the final CR LF (there can also be an EOF)
	csr	chIn
	jmp	cmd00		;return to main dispatch loop

code	setl	$
	org	data
words	dwm	0		;number of words in a record (one line)
type	dwm	0		;record type (00 or 01)
data	setl	$
	org	code

;-------------------------------------------------------------------------------
;G AAAA = Go execute at address.
; If no AAAA specified then resumes executing after breakpoint (eventually).

go	csr	hexin		;get starting address
	sta	(addr)

	lda	(flreg)		;restore flags saved at breakpoint
	psa
	pof
;	lds	(spreg)

	lda	(acreg)		;restore accumulator
	jmp	(addr)		;go to specified address

;-------------------------------------------------------------------------------
;S AAAA = Set memory contents.
; Enables one or more memory locations to be modified.
; e.g: MMMM nnn nn nnnn <Enter>
; Certain reserved memory locations can be changed to set the AC, SP and flags.

set	csr	hexin		;get address to start loading at
	sta	(addr)

set10	dsz	(emtyPtr)	;backup: check for cr (Enter key)
set20	csr	bufKey
	sub	" "
	joz	set20		;ignore space chars
	sub	0FFEDh		;= cr - " " (assembler no can handle)
	joz	cmd00		;return to main dispatch loop if cr

	dsz	(emtyPtr)	;else assume it's a hex char; backup
	csr	hexin		;get value to store into memory
	sta	[addr++]
	jmp	set10		;loop until cr is found

;-------------------------------------------------------------------------------
;B AAAA = Set (move) breakpoint.
; There is only one breakpoint. Moving it to location 0 (no AAAA) clears it.
; The 2 words at the specified address are saved, and a CSR xxxx is inserted.
; DMON is entered and the R command is executed when a breakpoint is executed.

break	lda	(opreg)		;restore instruction at previous breakpoint
	sta	[bpreg++]
	lda	(op2reg)
	sta	[bpreg]

	csr	hexin		;get new breakpoint location
	sta	(bpreg)

	lda	[bpreg]		;save instruction at new location
	sta	(opreg)

	lda	0317h		;store csr opcode in its place
	sta	[bpreg++]

	lda	[bpreg]
	sta	(op2reg)

	lda	brkhan		;address of breakpoint handler
	sta	[bpreg]

	dsz	(bpreg)		;undo ++
	 nop			;in case it skips
	jmp	cmd00		;return to main dispatch loop

;Breakpoint Handler: Come here when a breakpoint is executed.
brkhan	sta	(acreg)		;save accumulator
	psf			;save status flags
	poa
	sta	(flreg)
	poa			;get breakpoint address and balance stack
	dec
	dec
	sta	(pcreg)
				;(might want to change stacks)
	sts	(spreg)		;save stack pointer (at least for regs command)
	jmp	regs		;show registers then go to main dispatch loop

;===============================================================================
;                              INPUT ROUTINES
;-------------------------------------------------------------------------------
;Read in 4 ASCII hex characters and return their binary value in AC

hex4in	csr	hex2in
	csr	shl8
	sta	(regB)
	csr	hex2in
	ora	(regB)
	rtn

;-------------------------------------------------------------------------------
;Read in 2 ASCII hex characters and return their binary value in AC

hex2in	csr	hex1in
	csr	shl4
	sta	(regA)
	csr	hex1in
	ora	(regA)
	rtn

;-------------------------------------------------------------------------------
;Read in 1 ASCII hex character and return its binary value in AC

hex1in	csr	chIn
	sub	"0"
	psa
	sub	9+1
	poa
	joc	h1i20
	 sub	7
h1i20	rtn

code	setl	$
	org	data
regA	dwm	0
regB	dwm	0
data	setl	$
	org	code

;-------------------------------------------------------------------------------
;Input an ASCII hex number from the buffered keyboard. Convert the number to
; binary and return it in AC. The number must be terminated with a non-hex
; character (such as a CR). This routine skips any leading non-hex characters.
; If more than 4 hex characters are input then only the last 4 are used.

hexin	clr			;initialize
	sta	(regA)		;hex accumulator
	sta	(haveNum)	;flag
hi10	csr	bufKey
	sub	"0"		;convert ASCII hex character to binary
	psa
	sub	9+1
	poa
	joc	hi20		;jump if 0..9
	and	0FFDFh		;shift to uppercase
	sub	"A"-"0"
	sub	6+1		;(don't be tempted to combine subtracts)
	jnc	hi30		;jump if not hex char
	add	10+6+1		;A..F
hi20
	isz	(haveNum)	;indicate a hex character has been found
	sta	(temp)		;shift this hex digit into regA
	lda	(regA)
	csr	shl4
	ora	(temp)
	sta	(regA)
	jmp	hi10
hi30				;come here if non-hex character
	lda	(haveNum)	;has a hex character been processed?
	joz	hi10		;loop back if not
	lda	(regA)		;return with hex value in AC
	rtn

code	setl	$
	org	data
haveNum	dwm	0		;flag: a hex character has been read in
data	setl	$
	org	code

;===============================================================================
;                              OUTPUT ROUTINES
;-------------------------------------------------------------------------------
;Output a carriage return and linefeed.
; AC is destroyed.

crlf	lda	0Dh		;output a carriage return
	csr	chOut
	lda	0Ah		;and a linefeed
	jmp	chOut

;-------------------------------------------------------------------------------
;Output an ASCIIZ (zero-terminated) string at the address in AC.

textZ	psy
	may
txtz20	lda	(iy+0)		;output high byte first
	csr	lsr8
	joz	txtz90
	csr	chOut
	lda	(iy+0)
	iny
	and	00FFh
	joz	txtz90
	csr	chOut
	jmp	txtz20
txtz90	poy
	rtn

;-------------------------------------------------------------------------------
;Output the string pointed to by AC. Terminated with MSB on last character set.
; Two characters are packed per word, the way the DFB directive for the 
; assembler works (first byte in high 8 bits). AC is destroyed.

textP	psy
	may
txt20	lda	(iy+0)		;output high byte first (big-endian style)
	jmi	txt30		;jump if MSB is set (last char)
	csr	lsr8
	csr	chOut
	lda	(iy+0)
	and	0080h
	joz	txt25
	lda	(iy+0)
	jmp	txt35
txt25
	lda	(iy+0)
	iny
	and	007Fh
	csr	chOut
	jmp	txt20
txt30
	csr	lsr8
txt35	and	007Fh		;mask off MSB
	poy
	jmp	chOut		;output last character and return

;-------------------------------------------------------------------------------
;Output the value in AC as four ASCII hex digits

hexout
hex4out	psa
	csr	lsr8
	csr	hex2out
	poa
				;fall into hex2out...
;-------------------------------------------------------------------------------
;Output the low byte in AC as two ASCII hex digits

hex2out	psa
	csr	lsr4
	csr	hex1out
	poa
				;fall into hex1out...
;-------------------------------------------------------------------------------
;Output the low 4 bits in AC as an ASCII hex digit

hex1out	and	000Fh
	sub	10
	jmi	hd10
	 add	7
hd10	add	10+30h
	jmp	chOut		;go output character and return

;-------------------------------------------------------------------------------
;Output space character(s)

sp3out	csr	spOut		;output 3 space characters
sp2out	csr	spOut		;output 2 space characters
spOut	lda	" "		;output 1 space character
	jmp	chOut

;-------------------------------------------------------------------------------
;(Logical) Shift AC right xx bits

lsr15	lsr
lsr14	lsr
lsr13	lsr
lsr12	lsr
lsr11	lsr
lsr10	lsr
lsr9	lsr
lsr8	lsr
lsr7	lsr
lsr6	lsr
lsr5	lsr
lsr4	lsr
lsr3	lsr
lsr2	lsr
lsr1	lsr
lsr0	rtn

;-------------------------------------------------------------------------------
;Shift AC left xx bits

shl15	shl
shl14	shl
shl13	shl
shl12	shl
shl11	shl
shl10	shl
shl9	shl
shl8	shl
shl7	shl
shl6	shl
shl5	shl
shl4	shl
shl3	shl
shl2	shl
shl1	shl
shl0	rtn

;===============================================================================
;		     LOW-LEVEL CONSOLE I/O ROUTINES
;===============================================================================

code	setl	$		;switch to data section of memory
	org	data
;Table of routines to handle basic I/O functions
;dev number	  0	  1	  2	  3	  4	  5	  6	  7	  8
openIT	dwm	resBuf,	resKey,	dummIO,	dummIO,	dummIO,	dummIO,	dummIO,	dummIO,	psOpnI
openOT	dwm	dummIO,	dummIO,	dummIO,	dummIO,	dummIO,	dummIO,	dummIO,	dummIO,	psOpnO
chInT	dwm	bufKey,	keyIn,	dummIO,	dummIO,	dummIO,	dummIO,	dummIO,	dummIO,	psIn
chOutT	dwm	conOut,	tvOut,	prnOut,	dummIO,	dummIO,	dummIO,	dummIO,	dummIO,	psOut
closeT	dwm	dummIO,	dummIO,	dummIO,	dummIO,	dummIO,	dummIO,	dummIO,	dummIO,	dummIO
data	setl	$
	org	code

;-------------------------------------------------------------------------------
;Initialize input device number (nowDev).
;
openI	psa			;save registers
	psy

	lda	(nowDev)	;get device number
	may
	sub	9		;legal device number?
	jnc	ioErr		;(pjmp) jump if not

	lda	(iy+openIT)	;get address of routine
	sta	(opnI10-1)	;self-modifying code
	csr	0		;call routine corresponding to device
opnI10	joc	ioErr		;(pjmp) jump if error

	poy			;restore registers
	poa
	rtn

;-------------------------------------------------------------------------------
;Initialize output device number (nowDev).
;
openO	psa			;save registers
	psy

	lda	(nowDev)	;get device number
	may
	sub	9		;legal device number?
	jnc	ioErr		;(pjmp) jump if not

	lda	(iy+openOT)	;get address of routine
	sta	(opnO10-1)	;self-modifying code
	csr	0		;call routine corresponding to device
opnO10	joc	ioErr		;(pjmp) jump if error

	poy			;restore registers
	poa
	rtn

;-------------------------------------------------------------------------------
;Input byte from device number (nowDev), and return it in AC.
;
chIn	psa			;save registers (AC is saved in case of ioErr)
	psy

	lda	(nowDev)	;get device number
	may
	sub	9		;legal device number?
	jnc	ioErr		;(pjmp) jump if not

	lda	(iy+chInT)	;get address of routine
	sta	(chin10-1)	;self-modifying code
	csr	0		;call routine corresponding to device
chin10	joc	ioErrX		;(pjmp) jump if error

chin90	poy			;restore registers
	ins			;discard original AC value
	rtn

ioErrX	psa			;preserve AC (might be EOF)
	lda	3		;report RUN-TIME ERROR 3: I/O
	csr	error
	poa
	jmp	chin90		;exit

;-------------------------------------------------------------------------------
;Report I/O error.
; BEWARE OF STACK. This is only called from openI, openO, chIn, chOut, 
; and close. AC and IY are already pushed onto the stack.
;
ioErr	lda	3		;report RUN-TIME ERROR 3: I/O
	csr	error

	poy			;restore registers
	poa
	rtn

;-------------------------------------------------------------------------------
;Output byte in AC to device number (nowDev).
;
chOut	psa			;save registers
	psy

	lda	(nowDev)	;get device number
	may
	sub	9		;legal device number?
	jnc	ioErr		;(pjmp) jump if not

	lda	(iy+chOutT)	;get address of routine
	sta	(cho10-1)	;self-modifying code
	lda	(sp+1)		;get byte to output
	csr	0		;call routine corresponding to device
cho10	joc	ioErr		;(pjmp) jump if error

	poy			;restore registers
	poa
	rtn

;-------------------------------------------------------------------------------
;Close output device number (nowDev).
;
close	psa			;save registers
	psy

	lda	(nowDev)	;get device number
	may
	sub	9		;legal device number?
	jnc	ioErr		;(pjmp) jump if not

	lda	(iy+closeT)	;get address of routine
	sta	(clos10-1)	;self-modifying code
	csr	0		;call routine corresponding to device
clos10	joc	ioErr		;(pjmp) jump if error

	poy			;restore registers
	poa
	rtn

;===============================================================================
;Initialize the keyboard buffer to be empty (reset it).

resBuf	lda	kbBuf		;set both pointers to start of buffer
	sta	(fillPtr)
	sta	(emtyPtr)
dummIO	ccf			;indicate no errors
	rtn

;-------------------------------------------------------------------------------
;Get character from buffered keyboard and return it in AC. (Device 0.)
;

bufKey	lda	(emtyPtr)	;if empty pointer < fill pointer then there
	sub	(fillPtr)	; is at least one character in the buffer
	joc	ki40		;jump if so

;Get keystrokes and store characters into buffer until the Enter key is struck
	lda	kbBuf		;point to start of keyboard buffer
	sta	(fillPtr)
ki10	csr	keyIn		;get keystroke

	psa
	sub	bs		;is it a backspace?
	jnz	ki20		;jump if not

	lda	kbBuf		;if buf < ptr then [ptr--; echo]
	sub	(fillPtr)	;if buf - ptr < 0 then [ptr--; echo]
	poa
	jnc	ki10		;loop back for next char
	dsz	(fillPtr)
	csr	chOut		;echo the backspace
	jmp	ki10		;loop back for next character
ki20
	poa
	sta	[fillPtr++]	;store char into buffer

	psa			;check for buffer overflow
	sub	cr		;is char the Enter key (carriage return)?
	joz	ki30		;jump if so

	lda	(fillPtr)	;if ptr >= bufEnd then ptr--
	sub	kbBufEnd	;if ptr - bufEnd >= 0 then ptr-- else echo
	poa
	jnc	ki25
	csr	chOut		;echo the character
	jmp	ki10		;loop back for next character
ki25
	dsz	(fillPtr)	;point to last position in buffer
	jmp	ki10		;loop back for next character
ki30
	poa
	csr	chOut		;echo the carriage return
	lda	lf		;and add a line feed
	csr	chOut

	lda	kbBuf		;set empty pointer to start of buffer
	sta	(emtyPtr)

ki40	lda	[emtyPtr++]	;get character from buffer
	ccf			;no errors
	rtn			;and return it

code	setl	$
	org	data
kbBuf	dfs	80		;keyboard buffer (one char per word)
kbBufEnd equ	$		;last location of buffer +1
fillPtr	dwm	kbBuf		;pointer to fill keyboard buffer
emtyPtr	dwm	kbBuf		;pointer to empty it
data	setl	$
	org	code

;-------------------------------------------------------------------------------
;Output character in AC to the console (device 0).
;
conOut
	jmp	tvOut

;-------------------------------------------------------------------------------
;Clear keystrokes for device 1 input
;
resKey			;???
	ccf
	rtn

;-------------------------------------------------------------------------------
;Input a char from the keyboard (without echo) and return it in AC. (Device 1.)
;
keyIn
	if	sim
	 dwm	00FEh		;ChIn for simulator
	else
	 csr	Rcv		;receive character on serial port
	endi
	ccf			;indicate no errors
	rtn

;-------------------------------------------------------------------------------
;Output character in AC to the console, but with no ctrl-c, no tab handling
; etc. (device 1).
;
tvOut
	if	sim
	 dwm	00FDh		;CHOUT for simulator
	else
	 csr	Xmit		;transmit character on serial port
	endi
	ccf			;indicate no errors
	rtn

;-------------------------------------------------------------------------------
;Print character in AC (device 2).
;
prnOut
	jmp	tvOut

;===============================================================================
;Pseudo device handler (device 8).
; The pseudo device is a circular memory buffer which can be read from or
; written to. It is used to simplify string parsing and number conversion.
;
code	setl	$		;switch to data section of memory
	org	data
psIRed	dwm	0		;read pointer
psIWrt	dwm	0		;write pointer
psIStr	dwm	0		;start pointer
psuSiz	equ	256
psuBuf	dfs	psusiz		;buffer
data	setl	$
	org	code

;-------------------------------------------------------------------------------
;Open pseudo device for output.
;
psOpnO	psa			;save AC
	clr
	sta	(psIWrt)	;zero write pointer
	sta	(psIStr)	;zero start pointer
	poa			;restore AC
				;(pfall) also open input...
;-------------------------------------------------------------------------------
;Open pseudo device for input.
;
psOpnI	psa			;save AC
	lda	(psIStr)	;set read pointer to start
	sta	(psIRed)
	poa			;restore AC
	ccf			;indicate no errors
	rtn

;-------------------------------------------------------------------------------
;Input a byte from pseudo device and return it in AC.
;
psIn	psy

	lda	(psIRed)
	may
	sub	(psIWrt)	;read = write if empty
	jnz	psin1		;skip if more to read
	lda	eof		;signal end-of-file
	jmp	psin3
psin1
	ldy	(iy+psuBuf)	;read byte
	isz	(psIRed)	;advance pointer
	lda	(psIRed)
	sub	psuSiz		;wrap around?
	jnz	psin2		;skip if not
	sta	(psIRed)	;start over (psIRed:= 0)
psin2
	mya			;return byte in AC
psin3
	poy			;restore registers
	ccf			;no errors
	rtn

;-------------------------------------------------------------------------------
;Output the byte in AC to the pseudo device.

psout	psa			;save registers
	psy

	ldy	(psIWrt)	;get write pointer
	sta	(iy+psuBuf)	;store byte in buffer

;increment write pointer and handle wrap around and collisions
	isz	(psIWrt)	;increment write pointer
	lda	(psIWrt)	;wrap around?
	sub	psuSiz
	joc	incwt1		;skip if not
	clr			;else start over
	sta	(psIWrt)
incwt1
	lda	(psIWrt)	;caught start pointer?
	sub	(psIStr)
	jnz	incwt2		;skip if not

	isz	(psIStr)	;increment start pointer
	lda	(psIStr)	;wrap around?
	sub	psuSiz
	joc	incwt2		;skip if not
	clr			;else start over
	sta	(psIStr)
incwt2
	lda	(psIWrt)	;caught read pointer?
	sub	(psIRed)
	jnz	incwt3		;skip if not

	isz	(psIRed)	;increment read pointer
	lda	(psIRed)	;wrap around?
	sub	psuSiz
	joc	incwt3		;skip if not
	clr			;else start over
	sta	(psIRed)
incwt3
	poy			;restore registers
	poa
	ccf			;no errors
	rtn

;-------------------------------------------------------------------------------
;Common code to return from an (non-nested) procedure (or function).

retdo	pox			;restore heap pointer
	stx	(hp)
	pox			;restore IX
	rtn			;(functions return values thru AC)

;===============================================================================
