\D16SIMX.XPL	4-Aug-2005	Loren Blaney	loren_blaney@idcomm.com
\Simulator for John Doran's D16/M, version 2.1x
\WARNING: BRK, and [m+xx] and [(m++)+xx] address modes are not implemented
\ Also indexed JMP, Jcc, CSR, INP, OUT, ISZ and DSZ.

\REVISIONS:
\17-JAN-2002, Original
\19-FEB-2003, Modified to handle INT8 hex file output from cross assembler
\21-FEB-2003, Added CHIN (00FE) instruction for debugging
\25-FEB-2003, Changed name from D16 to SIM
\07-MAR-2003, Change load format from INT8 to MOT16; changed name from
\ SIM to D16SIM, CHOUT outputs to serial port
\11-MAR-2003, R command blanks display, for maximum speed.
\6-APR-2005, V2.0x: Added potential new instructions, such as "LDX (SP+3)".
\ Renamed from D16SIM to D16SIMX. Prevent loading beyond MemSize. CHOUT Scrolls.
\ Illegal opcode terminates RunMode.
\10-May-2005, V2.1x: Added dump and profile routines. (Hit Ctrl+D to stop run
\ mode so the P and D commands can be used.)
\4-Aug-2005, Corrected new (indexing) opcodes to match the hardware.

inc	\CXPL\CODESI;				\Intrinsics
ext	DoADD(addr Acc, addr Flags, Value),	\External assembly routines
	DoADC(addr Acc, addr Flags, Value),
	DoSUB(addr Acc, addr Flags, Value),
	DoSBC(addr Acc, addr Flags, Value);

\WARNING: This has only 24576 words of memory (instead of 65536 words)
def	MemSize = $6000,\Number of words of memory
	IOSize = $10;	\Number of 16-bit I/O ports

int	AC,		\Accumulator
	CsrX, CsrY,	\Cursor coordinates for special CHOUT pseudo op
	DumpAddr,	\Address where hex dump begins
	Key,		\Character from keyboard
	Flags,		\Bit 0 = carry; bit 1 = overflow
	IntEnab,	\Flag: Interrupt enabled
	RunMode,	\Flag: Run until HLT instruction
	Opcode,		\Instruction opcode
	Operand,	\Instruction operand
	IX		\Index register X
	IY		\Index register Y
	PC,		\Program counter
	Skip,		\Flag to skip next instruction (for ISZ, DSZ)
	SP;		\Stack pointer

int	Mem(MemSize),	\Memory, 16-bit
	IO(IOSize);	\Input/output, 16-bit

seg int Count(1);	\Count of number of times location was accessed

def	Nul=$00, Bel=$07, BS=$08, Tab=$09, LF=$0A, FF=$0C,	\Control chars
	CR=$0D, EOF=$1A, Esc=$1B, Space=$20, Ctrl=$40;

def	Black, Blue, Green, Cyan, Red, Magenta, Brown, White,	\Attribute colors
	Gray, LBlue, LGreen, LCyan, LRed, LMagenta, Yellow, BWhite; \ EGA palette



proc	Error(S);	\Display error message and wait for a keystroke
char	S;
int	I;
begin
Chout(0, Bel);
Text(0, S);
Crlf(0);
I:= Chin(1);
end;	\Error



func	CallInt(Int, AX, BX, CX, DX, BP, DS, ES); \Call software interrupt
int	Int, AX, BX, CX, DX, BP, DS, ES; \(Unused arguments need not be passed)
int	CpuReg;
begin
CpuReg:= Getreg;
CpuReg(0):= AX;
CpuReg(1):= BX;
CpuReg(2):= CX;
CpuReg(3):= DX;
CpuReg(6):= BP;
CpuReg(9):= DS;
CpuReg(11):= ES;
Softint(Int);
return CpuReg(0);		\return AX register
end;	\CallInt



func	LookKey;		\Returns next keystroke without reading it in
int	SC, Ch;			\ so it can be read in later (e.g: Intin(0))
begin
repeat until ChkKey;
SC:= CallInt($16, $0100);	\Function $01
Ch:= SC & $FF;
if Ch = 0 then Ch:= -(SC>>8);	\Return non-ASCII chars as negative scan code
return Ch;
end;	\LookKey



func	Hex1In;			\Read in one ASCII hex digit from disk file
int	Ch;			\ (for .HEX and 6502 .BIN files)
begin
Ch:= Chin(3);
if Ch>=^0 & Ch<=^9 then return Ch -^0;
if Ch>=^A & Ch<=^F then return Ch -^A +10;
return -1;			\Error--hex digit expected
end;	\Hex1In



func	Hex2In;			\Read in two ASCII hex digits from disk file
return Hex1In<<4 +Hex1In;



func	Hex4In;			\Read in four ASCII hex digits from disk file
return Hex2In<<8 +Hex2In;



func	DosOpen;
\Simple routine to open the input file specified on the command line.
\ The extension must be typed on the command line. No switches are allowed.
int	CpuReg,				\Register array from Getreg
	Handle;				\File handles
char	CmdTail;			\Copy of command tail
begin
CpuReg:= Getreg;			\Get DOS PSP and data segment
CmdTail:= Reserve($80);			\Get copy of command tail
Blit(CpuReg(11), $81, CpuReg(12), CmdTail, $7F);

Trap(false);
Handle:= Fopen(CmdTail, 0);		\Open file name for input
if GetErr then return false;
Trap(true);
Fset(Handle, ^I);
Openi(3);

return true;
end;	\DosOpen



proc	LEDOut(Reg);	\Display Reg as 16 LED's like this:
int	Reg;		\  * * * *  * * * *    *      *
int	I, J;
begin
Attrib(Red);
for J:= 0, 3 do
	begin
	for I:= 0, 3 do
		begin
		Chout(6, if Reg & $8000 then $0F else $FA);
		Reg:= Reg << 1;
		Chout(6, ^ );
		end;
	Chout(6, ^ );
	end;
end;	\LEDOut



proc	Dump;		\Hex dump 128 words (16 lines) starting at Addr
\0000:  0000 1111 2222 3333  4444 5555 6666 7777   ................
int	I, J, W;
char	Ch(16);
begin
for J:= 0, 8-1 do
	begin
	Hexout(0, DumpAddr);
	TEXT(0, ":  ");
	for I:= 0, 8-1 do
		begin
		W:= Mem(DumpAddr+I);
		Hexout(0, W);
		Chout(0, ^ );
		if (I&3) = 3 then Chout(0, ^ );
		Ch(I+I):= W>>8 & $7F;
		Ch(I+I+1):= W & $7F;
		end;
	Chout(0, ^ );
	Chout(0, ^ );
	for I:= 0, 16-1 do
		Chout(0, if Ch(I)<$20 ! Ch(I)=$7F then ^. else Ch(I));
	Crlf(0);
	DumpAddr:= DumpAddr + 8;
	end;
end;	\Dump



proc	Profile;	\Display 128 profile counters starting at Addr
\0000:	0	1	2	3	4	5	6	7
int	I, J, W;
char	Ch(16);
begin
for J:= 0, 8-1 do
	begin
	Hexout(0, DumpAddr);
	TEXT(0, ":  ");
	for I:= 0, 8-1 do
		begin
		W:= Count(0, DumpAddr+I);
		Intout(0, W);
		Chout(0, Tab);
		end;
	Crlf(0);
	DumpAddr:= DumpAddr + 8;
	end;
end;	\Profile



proc	Display;	\Display machine state (registers)
int	I, J, K, L;
begin
Cursor(0, 0);
TEXT(0, "		 D16/M Simulator, 2.1x

");
\ $XXXX   AC:	* * * *  * * * *    *      *
Text(0, "   ");
Hexout(0, AC);
Text(0, "   AC:	");
LEDOut(AC);
Crlf(0);

Text(0, "   ");
Hexout(0, IX);
Text(0, "   IX:	");
LEDOut(IX);
Crlf(0);

Text(0, "   ");
Hexout(0, IY);
Text(0, "   IY:	");
LEDOut(IY);
Crlf(0);

Text(0, "   ");
Hexout(0, SP);
Text(0, "   SP:	");
LEDOut(SP);
Crlf(0);

Text(0, "   ");
Hexout(0, PC);
Text(0, "   PC:	");
LEDOut(PC);
Crlf(0);

Text(0, "   ");
Hexout(0, Flags);
Text(0, "   CF:	");
Attrib(Red);
Chout(6, if Flags & $0001 then ^* else ^);
Text(0, "    OVF:   ");
Chout(6, if Flags & $0002 then ^* else ^);
Text(0, "    Int:   ");
Chout(6, if IntEnab then ^* else ^);

Crlf(0);
Crlf(0);

Attrib(White);
end;	\Display



proc	Execute;	\Execute an instruction
int	T,		\Temporary scratch
	Ind;		\Indirect address
begin
case Opcode of
 $001A:	AC:= AC + 1;						\INC
 $001B:	AC:= AC - 1;						\DEC
 $001C:	AC:= ~AC;						\COM
 $001D:	AC:= -AC;						\NEG
 $001E:	AC:= $FFFF;						\SET
 $001F:	AC:= 0;							\CLR
 $0020:	[Flags:= Flags & $FFFE;					\SHL
	if AC & $8000 then Flags:= Flags ! $0001;
	AC:= AC << 1];
 $0021:	[Flags:= Flags & $FFFE;					\LSR
	if AC & $0001 then Flags:= Flags ! $0001;
	AC:= AC >> 1];
 $0022:	[Flags:= Flags & $FFFE;					\ASR
	if AC & $0001 then Flags:= Flags ! $0001;
	AC:= AC >> 1;
	if AC & $4000 then AC:= AC ! $8000];
 $0023:	[T:= Flags;						\ROL
	Flags:= Flags & $FFFE;
	if AC & $8000 then Flags:= Flags ! $0001;
	AC:= AC<<1 ! T&$0001];
 $0024:	[T:= Flags;						\ROR
	Flags:= Flags&$FFFE ! AC&$0001;
	AC:= AC>>1 ! T<<15];
 $0025:	[Text(0, "$");						\LSW
	AC:= HexIn(0)];
 $0026:	Flags:= Flags ! $0001;					\SCF
 $0027:	Flags:= Flags & $FFFE;					\CCF
 $0028:	[SP:= SP - 1;						\PSA
	Mem(SP):= AC];
 $0029:	[AC:= Mem(SP);						\POA
	SP:= SP + 1];
 $002A:	[SP:= SP - 1;						\PSF
	Mem(SP):= Flags];
 $002B:	[Flags:= Mem(SP);					\POF
	SP:= SP + 1];
 $002C:	[IntEnab:= true];					\ENI
 $002D:	[IntEnab:= false];					\DII
 $002E:	[PC:= Mem(SP);						\RTN
	SP:= SP + 1];
 $002F:	[Flags:= Mem(SP);					\RTI
	SP:= SP + 1;
	AC:= Mem(SP);
	SP:= SP + 1;
	PC:= Mem(SP);
	SP:= SP + 1;
	IntEnab:= true];
 $0030:	[];							\NOP

\Potiential new instructions:
 $0035:	IX:= IX + 1;						\INX
 $0036:	IY:= IY + 1;						\INY
 $0037:	SP:= SP + 1;						\INS
 $0038:	IX:= IX - 1;						\DEX
 $0039:	IY:= IY - 1;						\DEY
 $003A:	SP:= SP - 1;						\DES
 $003B:	IX:= AC;						\MAX
 $003C:	IY:= AC;						\MAY
 $003D:	SP:= AC;						\MAS
 $003E:	AC:= IX;						\MXA
 $003F:	AC:= IY;						\MYA
 $0040:	AC:= SP;						\MSA
 $0041:	[SP:= SP - 1;						\PSX
	Mem(SP):= IX];
 $0042:	[IX:= Mem(SP);						\POX
	SP:= SP + 1];
 $0043:	[SP:= SP - 1;						\PSY
	Mem(SP):= IY];
 $0044:	[IY:= Mem(SP);						\POY
	SP:= SP + 1];
 $0045:	[Flags:= Flags & $FFFE;					\BSW
	if AC & $0100 then Flags:= Flags ! $0001;
	AC:= AC<<8 ! AC>>8];
 $0046:	[Flags:= Flags & $FFFE;					\BSL
	if AC & $0100 then Flags:= Flags ! $0001;
	AC:= AC << 8];
 $0047:	[Flags:= Flags & $FFFE;					\BSR
	if AC & $0080 then Flags:= Flags ! $0001;
	AC:= AC >> 8];
\ $0048:	[];							\\BRK

 $00FD:	begin							\CHOUT
	if RunMode then Chout(0, AC)
	else
		begin
		\Chout(4, AC);
		Cursor(CsrX, CsrY);
		case AC of
		  CR:	CsrX:= 0;
		  LF:	CsrY:= CsrY + 1
		other	begin
			Chout(6, AC);
			CsrX:= CsrX + 1;
			if CsrX > 79 then
				begin
				CsrX:= 0;
				CsrY:= CsrY+1;
				end;
			end;
		if CsrY > 24 then
			begin
			Cursor(0, 24);
			Chout(6, LF);	\scroll screen
			CsrY:= 24;
			end;
		end;
	end;

 $00FE:	begin
	AC:= Chin(1);						\CHIN
	if AC = $4\Ctrl+D\ then RunMode:= false;
	end;
 $00FF:	[OpenI(1);						\HLT
	T:= Chin(6);
	RunMode:= false]
other	begin
	Operand:= Mem(PC);
	Count(0, PC):= Count(0, PC) + 1;
	PC:= PC + 1;
	case Opcode of
	 $0301:	AC:= Operand;					\LDA
	 $0101:	AC:= Mem(Operand);
	 $0501:	[Ind:= Mem(Operand);
		AC:= Mem(Ind)];
	 $0D01:	[Ind:= Mem(Operand);
		AC:= Mem(Ind);
		Mem(Operand):= Ind+1];
	 $1101:	AC:= Mem(IX+Operand);
	 $2101:	AC:= Mem(IY+Operand);
	 $3101:	AC:= Mem(SP+Operand);

	 $0302:	SP:= Operand;					\LDS
	 $0102:	SP:= Mem(Operand);
	 $0502:	[Ind:= Mem(Operand);
		SP:= Mem(Ind)];
	 $0D02:	[Ind:= Mem(Operand);
		SP:= Mem(Ind);
		Mem(Operand):= Ind+1];
	 $1102:	SP:= Mem(IX+Operand);
	 $2102:	SP:= Mem(IY+Operand);
	 $3102:	SP:= Mem(SP+Operand);

	 $0331:	IX:= Operand;					\LDX
	 $0131:	IX:= Mem(Operand);
	 $0531:	[Ind:= Mem(Operand);
		IX:= Mem(Ind)];
	 $0D31:	[Ind:= Mem(Operand);
		IX:= Mem(Ind);
		Mem(Operand):= Ind+1];
	 $1131:	IX:= Mem(IX+Operand);
	 $2131:	IX:= Mem(IY+Operand);
	 $3131:	IX:= Mem(SP+Operand);

	 $0332:	IY:= Operand;					\LDY
	 $0132:	IY:= Mem(Operand);
	 $0532:	[Ind:= Mem(Operand);
		IY:= Mem(Ind)];
	 $0D32:	[Ind:= Mem(Operand);
		IY:= Mem(Ind);
		Mem(Operand):= Ind+1];
	 $1132:	IY:= Mem(IX+Operand);
	 $2132:	IY:= Mem(IY+Operand);
	 $3132:	IY:= Mem(SP+Operand);

	 $0103:	Mem(Operand):= AC;				\STA
	 $0503:	[Ind:= Mem(Operand);
		Mem(Ind):= AC];
	 $0D03:	[Ind:= Mem(Operand);
		Mem(Ind):= AC;
		Mem(Operand):= Ind+1];
	 $1103:	Mem(IX+Operand):= AC;
	 $2103:	Mem(IY+Operand):= AC;
	 $3103:	Mem(SP+Operand):= AC;

	 $0104:	Mem(Operand):= SP;				\STS
	 $0504:	[Ind:= Mem(Operand);
		Mem(Ind):= SP];
	 $0D04:	[Ind:= Mem(Operand);
		Mem(Ind):= SP;
		Mem(Operand):= Ind+1];
	 $1104:	Mem(IX+Operand):= SP;
	 $2104:	Mem(IY+Operand):= SP;
	 $3104:	Mem(SP+Operand):= SP;

	 $0133:	Mem(Operand):= IX;				\STX
	 $0533:	[Ind:= Mem(Operand);
		Mem(Ind):= IX];
	 $0D33:	[Ind:= Mem(Operand);
		Mem(Ind):= IX;
		Mem(Operand):= Ind+1];
	 $1133:	Mem(IX+Operand):= IX;
	 $2133:	Mem(IY+Operand):= IX;
	 $3133:	Mem(SP+Operand):= IX;

	 $0134:	Mem(Operand):= IY;				\STY
	 $0534:	[Ind:= Mem(Operand);
		Mem(Ind):= IY];
	 $0D34:	[Ind:= Mem(Operand);
		Mem(Ind):= IY;
		Mem(Operand):= Ind+1];
	 $1134:	Mem(IX+Operand):= IY;
	 $2134:	Mem(IY+Operand):= IY;
	 $3134:	Mem(SP+Operand):= IY;

	 $0305:	DoADD(addr AC, addr Flags, Operand);		\ADD
	 $0105:	DoADD(addr AC, addr Flags, Mem(Operand));
	 $0505:	[Ind:= Mem(Operand);
		DoADD(addr AC, addr Flags, Mem(Ind))];
	 $0D05:	[Ind:= Mem(Operand);
		DoADD(addr AC, addr Flags, Mem(Ind));
		Mem(Operand):= Ind+1];
	 $1105:	DoADD(addr AC, addr Flags, Mem(IX+Operand));
	 $2105:	DoADD(addr AC, addr Flags, Mem(IY+Operand));
	 $3105:	DoADD(addr AC, addr Flags, Mem(SP+Operand));

	 $0306:	DoADC(addr AC, addr Flags, Operand);		\ADC
	 $0106:	DoADC(addr AC, addr Flags, Mem(Operand));
	 $0506:	[Ind:= Mem(Operand);
		DoADC(addr AC, addr Flags, Mem(Ind))];
	 $0D06:	[Ind:= Mem(Operand);
		DoADC(addr AC, addr Flags, Mem(Ind));
		Mem(Operand):= Ind+1];
	 $1106:	DoADC(addr AC, addr Flags, Mem(IX+Operand));
	 $2106:	DoADC(addr AC, addr Flags, Mem(IY+Operand));
	 $3106:	DoADC(addr AC, addr Flags, Mem(SP+Operand));

	 $0307:	DoSUB(addr AC, addr Flags, Operand);		\SUB
	 $0107:	DoSUB(addr AC, addr Flags, Mem(Operand));
	 $0507:	[Ind:= Mem(Operand);
		DoSUB(addr AC, addr Flags, Mem(Ind))];
	 $0D07:	[Ind:= Mem(Operand);
		DoSUB(addr AC, addr Flags, Mem(Ind));
		Mem(Operand):= Ind+1];
	 $1107:	DoSUB(addr AC, addr Flags, Mem(IX+Operand));
	 $2107:	DoSUB(addr AC, addr Flags, Mem(IY+Operand));
	 $3107:	DoSUB(addr AC, addr Flags, Mem(SP+Operand));

	 $0308:	DoSBC(addr AC, addr Flags, Operand);		\SBC
	 $0108:	DoSBC(addr AC, addr Flags, Mem(Operand));
	 $0508:	[Ind:= Mem(Operand);
		DoSBC(addr AC, addr Flags, Mem(Ind))];
	 $0D08:	[Ind:= Mem(Operand);
		DoSBC(addr AC, addr Flags, Mem(Ind));
		Mem(Operand):= Ind+1];
	 $1108:	DoSBC(addr AC, addr Flags, Mem(IX+Operand));
	 $2108:	DoSBC(addr AC, addr Flags, Mem(IY+Operand));
	 $3108:	DoSBC(addr AC, addr Flags, Mem(SP+Operand));

	 $0309:	AC:= AC & Operand;				\AND
	 $0109:	AC:= AC & Mem(Operand);
	 $0509:	[Ind:= Mem(Operand);
		AC:= AC & Mem(Ind)];
	 $0D09:	[Ind:= Mem(Operand);
		AC:= AC & Mem(Ind);
		Mem(Operand):= Ind+1];
	 $1109:	AC:= AC & Mem(IX+Operand);
	 $2109:	AC:= AC & Mem(IY+Operand);
	 $3109:	AC:= AC & Mem(SP+Operand);

	 $030A:	AC:= AC ! Operand;				\ORA
	 $010A:	AC:= AC ! Mem(Operand);
	 $050A:	[Ind:= Mem(Operand);
		AC:= AC ! Mem(Ind)];
	 $0D0A:	[Ind:= Mem(Operand);
		AC:= AC ! Mem(Ind);
		Mem(Operand):= Ind+1];
	 $110A:	AC:= AC ! Mem(IX+Operand);
	 $210A:	AC:= AC ! Mem(IY+Operand);
	 $310A:	AC:= AC ! Mem(SP+Operand);

	 $030B:	AC:= AC | Operand;				\XOR
	 $010B:	AC:= AC | Mem(Operand);
	 $050B:	[Ind:= Mem(Operand);
		AC:= AC | Mem(Ind)];
	 $0D0B:	[Ind:= Mem(Operand);
		AC:= AC | Mem(Ind);
		Mem(Operand):= Ind+1];
	 $110B:	AC:= AC | Mem(IX+Operand);
	 $210B:	AC:= AC | Mem(IY+Operand);
	 $310B:	AC:= AC | Mem(SP+Operand);

	 $010C:	[Mem(Operand):= Mem(Operand) + 1;		\ISZ
		if Mem(Operand) = 0 then Skip:= true];
	 $050C:	[Ind:= Mem(Operand);
		Mem(Ind):= Mem(Ind) + 1;
		if Mem(Ind) = 0 then Skip:= true];

	 $010D:	[Mem(Operand):= Mem(Operand) - 1;		\DSZ
		if Mem(Operand) = 0 then Skip:= true];
	 $050D:	[Ind:= Mem(Operand);
		Mem(Ind):= Mem(Ind) - 1;
		if Mem(Ind) = 0 then Skip:= true];

	 $030E:	PC:= Operand;					\JMP
	 $010E:	PC:= Mem(Operand);

	 $030F:	if AC = 0 then PC:= Operand;			\JOZ
	 $010F:	if AC = 0 then PC:= Mem(Operand);

	 $0310:	if AC # 0 then PC:= Operand;			\JNZ
	 $0110:	if AC # 0 then PC:= Mem(Operand);

	 $0311:	if AC >= 0 then PC:= Operand;			\JPL
	 $0111:	if AC >= 0 then PC:= Mem(Operand);

	 $0312:	if AC < 0 then PC:= Operand;			\JMI
	 $0112:	if AC < 0 then PC:= Mem(Operand);

	 $0313:	if Flags & $0001 then PC:= Operand;		\JOC
	 $0113:	if Flags & $0001 then PC:= Mem(Operand);

	 $0314:	if (Flags & $0001) = 0 then PC:= Operand;	\JNC
	 $0114:	if (Flags & $0001) = 0 then PC:= Mem(Operand);

	 $0315:	if Flags & $0002 then PC:= Operand;		\JOV
	 $0115:	if Flags & $0002 then PC:= Mem(Operand);

	 $0316:	if (Flags & $0002) = 0 then PC:= Operand;	\JNV
	 $0116:	if (Flags & $0002) = 0 then PC:= Mem(Operand);

	 $0317:	[SP:= SP - 1;					\CSR
		Mem(SP):= PC;
		PC:= Operand];

	 $0118:	AC:= IO(Operand);				\INP
	 $0119:	IO(Operand):= AC				\OUT

	other	[Error("Illegal opcode");  RunMode:= false];
	end;
end;	\Execute



func	LoadRec;
\Load data for one MOT16 record, return 'false' when EOF record is read
int	Type,
	NumWords,
	ByteAddr,
	WordAddr,
	I;
begin
repeat until Chin(3) = ^S;	\Skip to start of record
Type:= Chin(3);			\Record type: 2=data; 8=EOF

NumWords:= Hex2In;		\Get the number of bytes in this record
\Subtract 3 for the load addr and 1 for the checksum. This gives the
\ number of data bytes. Divide by 2 to get the number of data words.
NumWords:= (NumWords-4) >>1;

\Get load address:
ByteAddr:= Hex2In;		\Discard first 2 hex chars. We can't load
				\ above MemSize anyway (= byte addr 2000)
ByteAddr:= Hex4In;		\Get byte load address
WordAddr:= ByteAddr >> 1;	\Convert it to a word load address

for I:= 0, NumWords-1 do	\load the data
	begin
	if WordAddr >= MemSize then [Error("File too big"); return false];
	Mem(WordAddr):= Hex4In;
	WordAddr:= WordAddr + 1;
	end;

\Ignore the checksum since it's virtually impossible to get a load error here.
\ (Over a serial link is another matter.)

return Type = ^2;
end;	\LoadRec



proc	Init;		\Initialize and load a program
int	I;
begin
RunMode:= false;
AC:= Ran($7FFF);
IX:= Ran($7FFF);
IY:= Ran($7FFF);
Flags:= Ran(4);
for I:= 0, MemSize-1 do Mem(I):= Ran($7FFF);
for I:= 0, IOSize-1 do IO(I):= Ran($7FFF);
SP:= Ran($7FFF);

\Load a program into memory
if not DosOpen then
	begin
	Text(0, "No input file specified
Usage: D16 filename.hex
");
	exit 1;
	end;
while LoadRec do;	\Load lines of data from hex file until EOF is found

PC:= 0;
Skip:= false;
IntEnab:= false;
CsrX:= 0;  CsrY:= \24\10;	\Output any text at bottom of screen

Count(0):= Malloc(MemSize>>3+1);
for I:= 0, MemSize-1 do Count(0, I):= 0;
end;	\Init



begin	\Main
Init;
Chout(0, FF);
CallInt($10, $0100, 0, $2000);		\turn off flashing cursor

SetWind(0, 0, 79, 24, 0, 0);		\enable scroll for 80x25 screen 

DumpAddr:= 0;
loop	begin
	if not RunMode then
		begin
		Display;
		Key:= Chin(6);	\Single step, Ctrl+C to exit (non-redirectable)
		case Key of
		  ^r,^R:[Runmode:= true;
			SetVid($03)];	\clear screen & restore flashing cursor
		  ^d,^D:begin
			if LookKey # CR then DumpAddr:= Hexin(0) else Crlf(0);
			Dump;
			end;
		  ^p,^P:begin
			if LookKey # CR then DumpAddr:= Hexin(0) else Crlf(0);
			Profile;
			end;
		  Esc:	quit
		other	[];
		end;

	Opcode:= Mem(PC);	\Fetch
	Count(0, PC):= Count(0, PC) + 1;
	PC:= PC + 1;

	if Skip then
		begin
		Skip:= false;
		\Handle the skip of the operand for 2-word instructions
		if Opcode > $00FF then PC:= PC + 1;
		end
	else	Execute;
	end;
SetVid($03);	\clear screen and restore flashing cursor
end;	\Main
